{-# OPTIONS -#include "Curses_hsc.h" #-}
{-# LINE 1 "Curses.hsc" #-}
module Curses (
{-# LINE 2 "Curses.hsc" #-}
    --------------------------------------------------------------------
    
    Window,      -- data Window deriving Eq
    stdScr,      -- :: Window
    initScr,     -- :: IO Window
    cBreak,      -- :: Bool -> IO ()
    raw,         -- :: Bool -> IO ()
    echo,        -- :: Bool -> IO ()
    nl,          -- :: Bool -> IO ()
    intrFlush,   -- :: Bool -> IO ()
    keypad,      -- :: Window -> Bool -> IO ()
    initCurses,  -- :: IO ()
    endWin,      -- :: IO ()
    
    --------------------------------------------------------------------
    
    scrSize,          -- :: IO (Int, Int)
    refresh,          -- :: IO ()
    setCursesOutConv, -- :: IO (Char -> [Byte]) -> IO ()
    
    --------------------------------------------------------------------
    
    hasColors,      -- :: IO Bool
    startColor,     -- :: IO ()
    Pair(..),       -- newtype Pair = Pair Int deriving (Eq, Ord, Ix)
    colorPairs,     -- :: IO Int
    Color(..),      -- newtype Color = Color Int deriving (Eq, Ord, Ix)
    colors,         -- :: IO Int
    black, red, green, yellow, blue, magenta, cyan, white, -- :: Color
    initPair,       -- :: Pair -> Color -> Color -> IO ()
    pairContent,    -- :: Pair -> IO (Color, Color)
    canChangeColor, -- :: IO Bool
    initColor,      -- :: Color -> (Int, Int, Int) -> IO ()
    colorContent,   -- :: Color -> IO (Int, Int, Int)
    
    --------------------------------------------------------------------
    
    Attr,  -- data Attr deriving Eq
    attr0, -- :: Attr
    
    isAltCharset, isBlink, isBold, isDim, isHorizontal, isInvis,
    isLeft, isLow, isProtect, isReverse, isRight, isStandout, isTop,
    isUnderline, isVertical,
        -- :: Attr -> Bool
    
    setAltCharset, setBlink, setBold, setDim, setHorizontal, setInvis,
    setLeft, setLow, setProtect, setReverse, setRight, setStandout,
    setTop, setUnderline, setVertical,
        -- :: Attr -> Bool -> Attr
    
    attrSet, -- :: Attr -> Pair -> IO ()
    attrOn, attrOff,
    
    --------------------------------------------------------------------

    ulCorner, llCorner, urCorner, lrCorner, rTee, lTee, bTee, tTee,
    hLine, vLine, plus, s1, s9, diamond, ckBoard, degree, plMinus,
    bullet, lArrow, rArrow, dArrow, uArrow, board, lantern, block,
    -- The rest of characters is not standard. ncurses says:
    -- "These aren't documented, but a lot of System Vs have them
    -- anyway (you can spot pprryyzz{{||}} in a lot of AT&T terminfo
    -- strings). The ACS_names may not match AT&T's, our source
    -- didn't know them."
    s3, s7, lEqual, gEqual, pi, nEqual, sterling,
        -- :: Char
    wAddStr, 
    addStr,        -- :: String -> IO ()
    addStrLn,      -- :: String -> IO ()
    addGraphStr,   -- :: String -> IO ()
    addGraphStrLn, -- :: String -> IO ()
    addLn,         -- :: IO ()
    
    --------------------------------------------------------------------
    
    bkgrndSet,      -- :: Attr -> Pair -> IO ()
    erase,          -- :: IO ()
    clrToEol,       -- :: IO ()
    move,           -- :: Int -> Int -> IO ()
    Visibility(..), -- data Visibility = Invisible | Visible | VeryVisible
    cursSet,        -- :: Visibility -> IO Visibility
    standout,standend,
    attrDim, attrBold,
    attrDimOn, attrDimOff,
    attrBoldOn, attrBoldOff,
    wAttrOn,
    wAttrOff, 
    touchWin,
    mvWAddStr,
    wMove,
    
    --------------------------------------------------------------------
    
    Key(..),
        -- data Key
        --     = KeyChar Char | KeyBreak | KeyDown | KeyUp | KeyLeft
        --     | KeyRight | KeyHome | KeyBackspace | KeyF Int | KeyDL
        --     | KeyIL | KeyDC | KeyIC | KeyEIC | KeyClear | KeyEOS
        --     | KeyEOL | KeySF | KeySR | KeyNPage | KeyPPage | KeySTab
        --     | KeyCTab | KeyCATab | KeyEnter | KeySReset | KeyReset
        --     | KeyPrint | KeyLL | KeyA1 | KeyA3 | KeyB2 | KeyC1
        --     | KeyC3 | KeyBTab | KeyBeg | KeyCancel | KeyClose
        --     | KeyCommand | KeyCopy | KeyCreate | KeyEnd | KeyExit
        --     | KeyFind | KeyHelp | KeyMark | KeyMessage | KeyMove
        --     | KeyNext | KeyOpen | KeyOptions | KeyPrevious | KeyRedo
        --     | KeyReference | KeyRefresh | KeyReplace | KeyRestart
        --     | KeyResume | KeySave | KeySBeg | KeySCancel
        --     | KeySCommand | KeySCopy | KeySCreate | KeySDC | KeySDL
        --     | KeySelect | KeySEnd | KeySEOL | KeySExit | KeySFind
        --     | KeySHelp | KeySHome | KeySIC | KeySLeft | KeySMessage
        --     | KeySMove | KeySNext | KeySOptions | KeySPrevious
        --     | KeySPrint | KeySRedo | KeySReplace | KeySRight
        --     | KeySRsume | KeySSave | KeySSuspend | KeySUndo
        --     | KeySuspend | KeyUndo | KeyUnknown Int
        --     deriving Eq
    
    getCh, getChHack,
    newPad, pRefresh, delWin, newWin,
    wClrToEol,
    def_prog_mode,
    reset_prog_mode,
    mySystem
    ) 
    
    --------------------------------------------------------------------
    where

import Prelude hiding (pi)
import Monad          (liftM)
import Char           (chr, ord)
import Ix             (Ix)
import Bits
import Concurrent

import Foreign
import CForeign

import GHC.IOBase

import UTF8


{-# LINE 143 "Curses.hsc" #-}

{-# LINE 144 "Curses.hsc" #-}

{-# LINE 147 "Curses.hsc" #-}


------------------------------------------------------------------------

fi = fromIntegral

throwIfErr :: Num a => String -> IO a -> IO a
throwIfErr name act = do
    res <- act
    if res == (-1)
{-# LINE 157 "Curses.hsc" #-}
        then ioError (userError ("Curses: "++name++" failed"))
        else return res

throwIfErr_ :: Num a => String -> IO a -> IO ()
throwIfErr_ name act = void $ throwIfErr name act

------------------------------------------------------------------------

data WindowTag = WindowTag
type Window = Ptr WindowTag

stdScr :: Window
stdScr = unsafePerformIO (peek stdscr)
-- #if !HAVE_FOREIGN_LABEL
foreign import ccall "hs_stdscr" stdscr :: Ptr Window

{-# LINE 173 "Curses.hsc" #-}
-- #else
-- foreign label stdscr :: Ptr Window
-- #endif

initScr :: IO Window
initScr = throwIfNull "initscr" initscr
foreign import ccall unsafe initscr :: IO Window

cBreak :: Bool -> IO ()
cBreak True  = throwIfErr_ "cbreak"   cbreak
cBreak False = throwIfErr_ "nocbreak" nocbreak
foreign import ccall unsafe cbreak :: IO CInt
foreign import ccall unsafe nocbreak :: IO CInt

raw :: Bool -> IO ()
raw False = throwIfErr_ "noraw" noraw
raw True  = throwIfErr_ "raw"   raw_c
foreign import ccall unsafe noraw :: IO CInt
foreign import ccall unsafe "raw" raw_c :: IO CInt

echo :: Bool -> IO ()
echo False = throwIfErr_ "noecho" noecho
echo True  = throwIfErr_ "echo"   echo_c
foreign import ccall unsafe noecho :: IO CInt
foreign import ccall unsafe "echo" echo_c :: IO CInt

nl :: Bool -> IO ()
nl True  = throwIfErr_ "nl"   nl_c
nl False = throwIfErr_ "nonl" nonl
foreign import ccall unsafe "nl" nl_c :: IO CInt
foreign import ccall unsafe nonl :: IO CInt

intrFlush :: Bool -> IO ()
intrFlush bf =
    throwIfErr_ "intrflush" $ intrflush stdScr (if bf then 1 else 0)
foreign import ccall unsafe intrflush :: Window -> (Int8) -> IO CInt
{-# LINE 209 "Curses.hsc" #-}

keypad :: Window -> Bool -> IO ()
keypad win bf =
    throwIfErr_ "keypad" $ keypad_c win (if bf then 1 else 0)
foreign import ccall unsafe "keypad" keypad_c :: Window -> (Int8) -> IO CInt
{-# LINE 214 "Curses.hsc" #-}

initCurses :: IO ()
initCurses = do
    initScr; cBreak True; echo False
    nl False; intrFlush False; keypad stdScr True

endWin :: IO ()
endWin = throwIfErr_ "endwin" endwin
foreign import ccall unsafe endwin :: IO CInt

------------------------------------------------------------------------

scrSize :: IO (Int, Int)
scrSize = do
    lines <- peek linesPtr
    cols  <- peek colsPtr
    return (fromIntegral lines, fromIntegral cols)

foreign import ccall unsafe "hs_lines" linesPtr :: Ptr CInt

{-# LINE 234 "Curses.hsc" #-}
foreign import ccall unsafe "hs_cols" colsPtr :: Ptr CInt

{-# LINE 236 "Curses.hsc" #-}

refresh :: IO ()
refresh = throwIfErr_ "refresh" refresh_c
foreign import ccall unsafe "refresh" refresh_c :: IO CInt

------------------------------------------------------------------------

hasColors :: IO Bool
hasColors = liftM (/= 0) has_colors
foreign import ccall unsafe has_colors :: IO (Int8)
{-# LINE 246 "Curses.hsc" #-}

startColor :: IO ()
startColor = throwIfErr_ "start_color" start_color
foreign import ccall unsafe start_color :: IO CInt

newtype Pair = Pair Int deriving (Eq, Ord, Ix)

colorPairs :: IO Int
colorPairs = fmap fromIntegral $ peek colorPairsPtr

foreign import ccall unsafe "hs_color_pairs" colorPairsPtr :: Ptr CInt

{-# LINE 258 "Curses.hsc" #-}

newtype Color = Color Int deriving (Eq, Ord, Ix)

colors :: IO Int
colors = liftM fromIntegral $ peek colorsPtr
foreign import ccall unsafe "hs_colors" colorsPtr :: Ptr CInt

{-# LINE 265 "Curses.hsc" #-}

black, red, green, yellow, blue, magenta, cyan, white :: Color
black   = Color (0)
{-# LINE 268 "Curses.hsc" #-}
red     = Color (1)
{-# LINE 269 "Curses.hsc" #-}
green   = Color (2)
{-# LINE 270 "Curses.hsc" #-}
yellow  = Color (3)
{-# LINE 271 "Curses.hsc" #-}
blue    = Color (4)
{-# LINE 272 "Curses.hsc" #-}
magenta = Color (5)
{-# LINE 273 "Curses.hsc" #-}
cyan    = Color (6)
{-# LINE 274 "Curses.hsc" #-}
white   = Color (7)
{-# LINE 275 "Curses.hsc" #-}

initPair :: Pair -> Color -> Color -> IO ()
initPair (Pair p) (Color f) (Color b) =
    throwIfErr_ "init_pair" $
        init_pair (fromIntegral p) (fromIntegral f) (fromIntegral b)
foreign import ccall unsafe init_pair :: CShort -> CShort -> CShort -> IO CInt

pairContent :: Pair -> IO (Color, Color)
pairContent (Pair p) =
    alloca $ \fPtr ->
    alloca $ \bPtr -> do
        throwIfErr "pair_content" $ pair_content (fromIntegral p) fPtr bPtr
        f <- peek fPtr
        b <- peek bPtr
        return (Color (fromIntegral f), Color (fromIntegral b))
foreign import ccall unsafe pair_content :: CShort -> Ptr CShort -> Ptr CShort -> IO CInt

canChangeColor :: IO Bool
canChangeColor = liftM (/= 0) can_change_color
foreign import ccall unsafe can_change_color :: IO (Int8)
{-# LINE 295 "Curses.hsc" #-}

initColor :: Color -> (Int, Int, Int) -> IO ()
initColor (Color c) (r, g, b) =
    throwIfErr_ "init_color" $
        init_color (fromIntegral c) (fromIntegral r) (fromIntegral g) (fromIntegral b)
foreign import ccall unsafe init_color :: CShort -> CShort -> CShort -> CShort -> IO CInt

colorContent :: Color -> IO (Int, Int, Int)
colorContent (Color c) =
    alloca $ \rPtr ->
    alloca $ \gPtr ->
    alloca $ \bPtr -> do
        throwIfErr "color_content" $ color_content (fromIntegral c) rPtr gPtr bPtr
        r <- peek rPtr
        g <- peek gPtr
        b <- peek bPtr
        return (fromIntegral r, fromIntegral g, fromIntegral b)
foreign import ccall unsafe color_content :: CShort -> Ptr CShort -> Ptr CShort -> Ptr CShort -> IO CInt

foreign import ccall unsafe "hs_curses_color_pair" colorPair :: Pair -> (Word32)
{-# LINE 315 "Curses.hsc" #-}

{-# LINE 316 "Curses.hsc" #-}

------------------------------------------------------------------------

newtype Attr = Attr (Word32) deriving Eq
{-# LINE 320 "Curses.hsc" #-}

attr0 :: Attr
attr0 = Attr (0)
{-# LINE 323 "Curses.hsc" #-}

isAltCharset, isBlink, isBold, isDim, isHorizontal, isInvis, isLeft,
    isLow, isProtect, isReverse, isRight, isStandout, isTop,
    isUnderline, isVertical
    :: Attr -> Bool
isAltCharset = isAttr (4194304)
{-# LINE 329 "Curses.hsc" #-}
isBlink      = isAttr (524288)
{-# LINE 330 "Curses.hsc" #-}
isBold       = isAttr (2097152)
{-# LINE 331 "Curses.hsc" #-}
isDim        = isAttr (1048576)
{-# LINE 332 "Curses.hsc" #-}
isHorizontal = isAttr (33554432)
{-# LINE 333 "Curses.hsc" #-}
isInvis      = isAttr (8388608)
{-# LINE 334 "Curses.hsc" #-}
isLeft       = isAttr (67108864)
{-# LINE 335 "Curses.hsc" #-}
isLow        = isAttr (134217728)
{-# LINE 336 "Curses.hsc" #-}
isProtect    = isAttr (16777216)
{-# LINE 337 "Curses.hsc" #-}
isReverse    = isAttr (262144)
{-# LINE 338 "Curses.hsc" #-}
isRight      = isAttr (268435456)
{-# LINE 339 "Curses.hsc" #-}
isStandout   = isAttr (65536)
{-# LINE 340 "Curses.hsc" #-}
isTop        = isAttr (536870912)
{-# LINE 341 "Curses.hsc" #-}
isUnderline  = isAttr (131072)
{-# LINE 342 "Curses.hsc" #-}
isVertical   = isAttr (1073741824)
{-# LINE 343 "Curses.hsc" #-}

isAttr :: (Word32) -> Attr -> Bool
{-# LINE 345 "Curses.hsc" #-}
isAttr bit (Attr a) = a .&. bit /= 0

setAltCharset, setBlink, setBold, setDim, setHorizontal, setInvis,
    setLeft, setLow, setProtect, setReverse, setRight, setStandout,
    setTop, setUnderline, setVertical
    :: Attr -> Bool -> Attr
setAltCharset = setAttr (4194304)
{-# LINE 352 "Curses.hsc" #-}
setBlink      = setAttr (524288)
{-# LINE 353 "Curses.hsc" #-}
setBold       = setAttr (2097152)
{-# LINE 354 "Curses.hsc" #-}
setDim        = setAttr (1048576)
{-# LINE 355 "Curses.hsc" #-}
setHorizontal = setAttr (33554432)
{-# LINE 356 "Curses.hsc" #-}
setInvis      = setAttr (8388608)
{-# LINE 357 "Curses.hsc" #-}
setLeft       = setAttr (67108864)
{-# LINE 358 "Curses.hsc" #-}
setLow        = setAttr (134217728)
{-# LINE 359 "Curses.hsc" #-}
setProtect    = setAttr (16777216)
{-# LINE 360 "Curses.hsc" #-}
setReverse    = setAttr (262144)
{-# LINE 361 "Curses.hsc" #-}
setRight      = setAttr (268435456)
{-# LINE 362 "Curses.hsc" #-}
setStandout   = setAttr (65536)
{-# LINE 363 "Curses.hsc" #-}
setTop        = setAttr (536870912)
{-# LINE 364 "Curses.hsc" #-}
setUnderline  = setAttr (131072)
{-# LINE 365 "Curses.hsc" #-}
setVertical   = setAttr (1073741824)
{-# LINE 366 "Curses.hsc" #-}

setAttr :: (Word32) -> Attr -> Bool -> Attr
{-# LINE 368 "Curses.hsc" #-}
setAttr bit (Attr a) False = Attr (a .&. complement bit)
setAttr bit (Attr a) True  = Attr (a .|.            bit)

attrSet :: Attr -> Pair -> IO ()
attrSet attr (Pair p) = throwIfErr_ "attrset" $
    attr_set attr (fromIntegral p) nullPtr
foreign import ccall unsafe attr_set :: Attr -> CShort -> Ptr () -> IO Int

attrOn :: Attr -> IO ()
attrOn (Attr attr) = throwIfErr_ "attr_on" $
    attr_on attr nullPtr
foreign import ccall "attr_on" attr_on :: (Word32) -> Ptr a -> IO Int
{-# LINE 380 "Curses.hsc" #-}


attrOff :: Attr -> IO ()
attrOff (Attr attr) = throwIfErr_ "attr_off" $
    attr_off attr nullPtr
foreign import ccall "attr_off" attr_off :: (Word32) -> Ptr a -> IO Int
{-# LINE 386 "Curses.hsc" #-}

foreign import ccall "attron" attron :: Int -> IO Int
foreign import ccall "attroff" attroff :: Int -> IO Int

foreign import ccall unsafe wattron :: Window -> CInt -> IO CInt
foreign import ccall unsafe wattroff :: Window -> CInt -> IO CInt

wAttrOn :: Window -> Int -> IO ()
wAttrOn w x = throwIfErr_ "wattron" $ wattron w (fi x)
wAttrOff :: Window -> Int -> IO ()
wAttrOff w x = throwIfErr_ "wattroff" $ wattroff w (fi x)

attrDimOn :: IO ()
attrDimOn  = throwIfErr_ "attron A_DIM" $
    attron (1048576) 
{-# LINE 401 "Curses.hsc" #-}

attrDimOff :: IO ()
attrDimOff = throwIfErr_ "attroff A_DIM" $
    attroff (1048576) 
{-# LINE 405 "Curses.hsc" #-}

attrBoldOn :: IO ()
attrBoldOn  = throwIfErr_ "attron A_BOLD" $
    attron (2097152) 
{-# LINE 409 "Curses.hsc" #-}

attrBoldOff :: IO ()
attrBoldOff = throwIfErr_ "attroff A_BOLD" $
    attroff (2097152) 
{-# LINE 413 "Curses.hsc" #-}

foreign import ccall standout :: IO Int
foreign import ccall standend :: IO Int

attrDim :: Int
attrDim = (1048576)
{-# LINE 419 "Curses.hsc" #-}
attrBold :: Int
attrBold = (2097152)
{-# LINE 421 "Curses.hsc" #-}

------------------------------------------------------------------------


{-# LINE 612 "Curses.hsc" #-}

ulCorner, llCorner, urCorner, lrCorner, rTee, lTee, bTee, tTee, hLine,
    vLine, plus, s1, s9, diamond, ckBoard, degree, plMinus, bullet,
    lArrow, rArrow, dArrow, uArrow, board, lantern, block,
    s3, s7, lEqual, gEqual, pi, nEqual, sterling
    :: Char

ulCorner = '+'
llCorner = '+'
urCorner = '+'
lrCorner = '+'
rTee     = '+'
lTee     = '+'
bTee     = '+'
tTee     = '+'
hLine    = '-'
vLine    = '|'
plus     = '+'
s1       = '-'
s9       = '_'
diamond  = '+'
ckBoard  = ':'
degree   = '\''
plMinus  = '#'
bullet   = 'o'
lArrow   = '<'
rArrow   = '>'
dArrow   = 'v'
uArrow   = '^'
board    = '#'
lantern  = '#'
block    = '#'
s3       = '-'
s7       = '-'
lEqual   = '<'
gEqual   = '>'
pi       = '*'
nEqual   = '!'
sterling = 'f'


setCursesOutConv :: ([Char] -> [Word8]) -> IO ()
setCursesOutConv newConv = writeIORef cursesOutConv (withCConvString newConv)

{-# NOINLINE cursesOutConv #-}
cursesOutConv :: IORef (String -> (Ptr CChar -> IO b) -> IO b)
cursesOutConv = unsafePerformIO $ newIORef withCString

withCUTFString :: String -> (Ptr a -> IO b) -> IO b
withCUTFString s f = withArray0 0 (toUTF s) (f . castPtr)

withCConvString :: ([Char] -> [Word8]) -> String -> (Ptr CChar -> IO b) -> IO b
withCConvString conv s f = withArray0 0 (conv s) (f . castPtr)

addStr :: String -> IO ()
addStr str = do
	conv <- readIORef cursesOutConv
	throwIfErr_ "addstr" $ conv str addstr
foreign import ccall unsafe addstr :: Ptr CChar -> IO CInt

wAddStr :: Window -> String -> IO ()
wAddStr w str = do
	conv <- readIORef cursesOutConv
	throwIfErr_ "waddstr" $ conv str (waddstr w)

mvWAddStr :: Window -> Int -> Int -> String -> IO ()
mvWAddStr w y x str = do
	conv <- readIORef cursesOutConv
	throwIfErr_ "mvwaddstr" $ conv str (mvwaddstr w (fi y) (fi x))

foreign import ccall unsafe waddstr :: Window -> Ptr CChar -> IO CInt
foreign import ccall unsafe mvwaddstr :: Window -> CInt -> CInt -> Ptr CChar -> IO CInt

addStrLn :: String -> IO ()
addStrLn str = do addStr str; addLn

addGraphStr :: String -> IO ()
addGraphStr = addStr

addGraphStrLn :: String -> IO ()
addGraphStrLn = addStrLn


{-# LINE 695 "Curses.hsc" #-}

addLn :: IO ()
addLn = throwIfErr_ "addch" $ addch (fromIntegral (ord '\n'))

foreign import ccall unsafe addch :: (Word32) -> IO CInt
{-# LINE 700 "Curses.hsc" #-}

------------------------------------------------------------------------


{-# LINE 706 "Curses.hsc" #-}

bkgrndSet :: Attr -> Pair -> IO ()
bkgrndSet (Attr a) p = bkgdset $
    fromIntegral (ord ' ') .|.
    (if a .&. 4194304 /= 0 then 4194304 else 0) .|.
{-# LINE 711 "Curses.hsc" #-}
    (if a .&. 524288 /= 0 then 524288 else 0) .|.
{-# LINE 712 "Curses.hsc" #-}
    (if a .&. 2097152 /= 0 then 2097152 else 0) .|.
{-# LINE 713 "Curses.hsc" #-}
    (if a .&. 1048576 /= 0 then 1048576 else 0) .|.
{-# LINE 714 "Curses.hsc" #-}
    (if a .&. 8388608 /= 0 then 8388608 else 0) .|.
{-# LINE 715 "Curses.hsc" #-}
    (if a .&. 16777216 /= 0 then 16777216 else 0) .|.
{-# LINE 716 "Curses.hsc" #-}
    (if a .&. 262144 /= 0 then 262144 else 0) .|.
{-# LINE 717 "Curses.hsc" #-}
    (if a .&. 65536 /= 0 then 65536 else 0) .|.
{-# LINE 718 "Curses.hsc" #-}
    (if a .&. 131072 /= 0 then 131072 else 0) .|.
{-# LINE 719 "Curses.hsc" #-}
    colorPair p
foreign import ccall unsafe bkgdset :: (Word32) -> IO ()
{-# LINE 721 "Curses.hsc" #-}

erase :: IO ()
erase = throwIfErr_ "erase" erase_c
foreign import ccall unsafe "erase" erase_c :: IO CInt

clrToEol :: IO ()
clrToEol = throwIfErr_ "clrtoeol" clrtoeol
foreign import ccall unsafe clrtoeol :: IO CInt

move :: Int -> Int -> IO ()
move y x =
    throwIfErr_ "move" $ move_c (fromIntegral y) (fromIntegral x)
foreign import ccall unsafe "move" move_c :: CInt -> CInt -> IO CInt

wMove :: Window -> Int -> Int -> IO ()
wMove w y x = throwIfErr_ "wmove" $ wmove w (fi y) (fi x)
foreign import ccall unsafe  wmove :: Window -> CInt -> CInt -> IO CInt

data Visibility = Invisible | Visible | VeryVisible

cursSet :: Visibility -> IO Visibility
cursSet vis = do
    oldVis_c <- throwIfErr "curs_set" $ curs_set vis_c
    case toIntIfNeeded oldVis_c of
        0 -> return Invisible
        1 -> return Visible
        2 -> return VeryVisible
        _ -> ioError $ userError $
            "Curses: Unexpected curs_set result ("++show oldVis_c++")"
    where
    vis_c :: CInt
    vis_c = case vis of
        Invisible   -> 0
        Visible     -> 1
        VeryVisible -> 2
foreign import ccall unsafe curs_set :: CInt -> IO CInt


------------------------------------------------------------------------

toIntIfNeeded = id

data Key
    = KeyChar Char | KeyBreak | KeyDown | KeyUp | KeyLeft | KeyRight
    | KeyHome | KeyBackspace | KeyF Int | KeyDL | KeyIL | KeyDC
    | KeyIC | KeyEIC | KeyClear | KeyEOS | KeyEOL | KeySF | KeySR
    | KeyNPage | KeyPPage | KeySTab | KeyCTab | KeyCATab | KeyEnter
    | KeySReset | KeyReset | KeyPrint | KeyLL | KeyA1 | KeyA3
    | KeyB2 | KeyC1 | KeyC3 | KeyBTab | KeyBeg | KeyCancel | KeyClose
    | KeyCommand | KeyCopy | KeyCreate | KeyEnd | KeyExit | KeyFind
    | KeyHelp | KeyMark | KeyMessage | KeyMove | KeyNext | KeyOpen
    | KeyOptions | KeyPrevious | KeyRedo | KeyReference | KeyRefresh
    | KeyReplace | KeyRestart | KeyResume | KeySave | KeySBeg
    | KeySCancel | KeySCommand | KeySCopy | KeySCreate | KeySDC
    | KeySDL | KeySelect | KeySEnd | KeySEOL | KeySExit | KeySFind
    | KeySHelp | KeySHome | KeySIC | KeySLeft | KeySMessage | KeySMove
    | KeySNext | KeySOptions | KeySPrevious | KeySPrint | KeySRedo
    | KeySReplace | KeySRight | KeySRsume | KeySSave | KeySSuspend
    | KeySUndo | KeySuspend | KeyUndo | KeyResize | KeyUnknown Int
    deriving (Eq,Show)

decodeKey :: CInt -> Key
decodeKey key = case toIntIfNeeded key of
    _ | key >= 0 && key <= 255 -> KeyChar (chr (fromIntegral key))
    (257)         -> KeyBreak
{-# LINE 786 "Curses.hsc" #-}
    (258)          -> KeyDown
{-# LINE 787 "Curses.hsc" #-}
    (259)            -> KeyUp
{-# LINE 788 "Curses.hsc" #-}
    (260)          -> KeyLeft
{-# LINE 789 "Curses.hsc" #-}
    (261)         -> KeyRight
{-# LINE 790 "Curses.hsc" #-}
    (262)          -> KeyHome
{-# LINE 791 "Curses.hsc" #-}
    (263)     -> KeyBackspace
{-# LINE 792 "Curses.hsc" #-}
    _ | key >= (264) && key <= (327)
{-# LINE 793 "Curses.hsc" #-}
                               -> KeyF (fromIntegral (key - 264))
{-# LINE 794 "Curses.hsc" #-}
    (328)            -> KeyDL
{-# LINE 795 "Curses.hsc" #-}
    (329)            -> KeyIL
{-# LINE 796 "Curses.hsc" #-}
    (330)            -> KeyDC
{-# LINE 797 "Curses.hsc" #-}
    (331)            -> KeyIC
{-# LINE 798 "Curses.hsc" #-}
    (332)           -> KeyEIC
{-# LINE 799 "Curses.hsc" #-}
    (333)         -> KeyClear
{-# LINE 800 "Curses.hsc" #-}
    (334)           -> KeyEOS
{-# LINE 801 "Curses.hsc" #-}
    (335)           -> KeyEOL
{-# LINE 802 "Curses.hsc" #-}
    (336)            -> KeySF
{-# LINE 803 "Curses.hsc" #-}
    (337)            -> KeySR
{-# LINE 804 "Curses.hsc" #-}
    (338)         -> KeyNPage
{-# LINE 805 "Curses.hsc" #-}
    (339)         -> KeyPPage
{-# LINE 806 "Curses.hsc" #-}
    (340)          -> KeySTab
{-# LINE 807 "Curses.hsc" #-}
    (341)          -> KeyCTab
{-# LINE 808 "Curses.hsc" #-}
    (342)         -> KeyCATab
{-# LINE 809 "Curses.hsc" #-}
    (343)         -> KeyEnter
{-# LINE 810 "Curses.hsc" #-}
    (344)        -> KeySReset
{-# LINE 811 "Curses.hsc" #-}
    (345)         -> KeyReset
{-# LINE 812 "Curses.hsc" #-}
    (346)         -> KeyPrint
{-# LINE 813 "Curses.hsc" #-}
    (347)            -> KeyLL
{-# LINE 814 "Curses.hsc" #-}
    (348)            -> KeyA1
{-# LINE 815 "Curses.hsc" #-}
    (349)            -> KeyA3
{-# LINE 816 "Curses.hsc" #-}
    (350)            -> KeyB2
{-# LINE 817 "Curses.hsc" #-}
    (351)            -> KeyC1
{-# LINE 818 "Curses.hsc" #-}
    (352)            -> KeyC3
{-# LINE 819 "Curses.hsc" #-}
    (353)          -> KeyBTab
{-# LINE 820 "Curses.hsc" #-}
    (354)           -> KeyBeg
{-# LINE 821 "Curses.hsc" #-}
    (355)        -> KeyCancel
{-# LINE 822 "Curses.hsc" #-}
    (356)         -> KeyClose
{-# LINE 823 "Curses.hsc" #-}
    (357)       -> KeyCommand
{-# LINE 824 "Curses.hsc" #-}
    (358)          -> KeyCopy
{-# LINE 825 "Curses.hsc" #-}
    (359)        -> KeyCreate
{-# LINE 826 "Curses.hsc" #-}
    (360)           -> KeyEnd
{-# LINE 827 "Curses.hsc" #-}
    (361)          -> KeyExit
{-# LINE 828 "Curses.hsc" #-}
    (362)          -> KeyFind
{-# LINE 829 "Curses.hsc" #-}
    (363)          -> KeyHelp
{-# LINE 830 "Curses.hsc" #-}
    (364)          -> KeyMark
{-# LINE 831 "Curses.hsc" #-}
    (365)       -> KeyMessage
{-# LINE 832 "Curses.hsc" #-}
    (366)          -> KeyMove
{-# LINE 833 "Curses.hsc" #-}
    (367)          -> KeyNext
{-# LINE 834 "Curses.hsc" #-}
    (368)          -> KeyOpen
{-# LINE 835 "Curses.hsc" #-}
    (369)       -> KeyOptions
{-# LINE 836 "Curses.hsc" #-}
    (370)      -> KeyPrevious
{-# LINE 837 "Curses.hsc" #-}
    (371)          -> KeyRedo
{-# LINE 838 "Curses.hsc" #-}
    (372)     -> KeyReference
{-# LINE 839 "Curses.hsc" #-}
    (373)       -> KeyRefresh
{-# LINE 840 "Curses.hsc" #-}
    (374)       -> KeyReplace
{-# LINE 841 "Curses.hsc" #-}
    (375)       -> KeyRestart
{-# LINE 842 "Curses.hsc" #-}
    (376)        -> KeyResume
{-# LINE 843 "Curses.hsc" #-}
    (377)          -> KeySave
{-# LINE 844 "Curses.hsc" #-}
    (378)          -> KeySBeg
{-# LINE 845 "Curses.hsc" #-}
    (379)       -> KeySCancel
{-# LINE 846 "Curses.hsc" #-}
    (380)      -> KeySCommand
{-# LINE 847 "Curses.hsc" #-}
    (381)         -> KeySCopy
{-# LINE 848 "Curses.hsc" #-}
    (382)       -> KeySCreate
{-# LINE 849 "Curses.hsc" #-}
    (383)           -> KeySDC
{-# LINE 850 "Curses.hsc" #-}
    (384)           -> KeySDL
{-# LINE 851 "Curses.hsc" #-}
    (385)        -> KeySelect
{-# LINE 852 "Curses.hsc" #-}
    (386)          -> KeySEnd
{-# LINE 853 "Curses.hsc" #-}
    (387)          -> KeySEOL
{-# LINE 854 "Curses.hsc" #-}
    (388)         -> KeySExit
{-# LINE 855 "Curses.hsc" #-}
    (389)         -> KeySFind
{-# LINE 856 "Curses.hsc" #-}
    (390)         -> KeySHelp
{-# LINE 857 "Curses.hsc" #-}
    (391)         -> KeySHome
{-# LINE 858 "Curses.hsc" #-}
    (392)           -> KeySIC
{-# LINE 859 "Curses.hsc" #-}
    (393)         -> KeySLeft
{-# LINE 860 "Curses.hsc" #-}
    (394)      -> KeySMessage
{-# LINE 861 "Curses.hsc" #-}
    (395)         -> KeySMove
{-# LINE 862 "Curses.hsc" #-}
    (396)         -> KeySNext
{-# LINE 863 "Curses.hsc" #-}
    (397)      -> KeySOptions
{-# LINE 864 "Curses.hsc" #-}
    (398)     -> KeySPrevious
{-# LINE 865 "Curses.hsc" #-}
    (399)        -> KeySPrint
{-# LINE 866 "Curses.hsc" #-}
    (400)         -> KeySRedo
{-# LINE 867 "Curses.hsc" #-}
    (401)      -> KeySReplace
{-# LINE 868 "Curses.hsc" #-}
    (402)        -> KeySRight
{-# LINE 869 "Curses.hsc" #-}
    (403)        -> KeySRsume
{-# LINE 870 "Curses.hsc" #-}
    (404)         -> KeySSave
{-# LINE 871 "Curses.hsc" #-}
    (405)      -> KeySSuspend
{-# LINE 872 "Curses.hsc" #-}
    (406)         -> KeySUndo
{-# LINE 873 "Curses.hsc" #-}
    (407)       -> KeySuspend
{-# LINE 874 "Curses.hsc" #-}
    (408)          -> KeyUndo
{-# LINE 875 "Curses.hsc" #-}
    (410)        -> KeyResize
{-# LINE 876 "Curses.hsc" #-}
    _                          -> KeyUnknown (fromIntegral key)




getChHack :: IO Key
getChHack = liftM (decodeKey . fromIntegral . ord) getChar

touchWin :: Window -> IO ()
touchWin w = throwIfErr_ "touchwin" $ touchwin w
foreign import ccall touchwin :: Window -> IO CInt

newPad :: Int -> Int -> IO Window
newPad nlines ncols = throwIfNull "newpad" $ newpad (fromIntegral nlines) (fromIntegral ncols)

pRefresh :: Window -> Int -> Int -> Int -> Int -> Int -> Int -> IO ()
pRefresh pad pminrow pmincol sminrow smincol smaxrow smaxcol = throwIfErr_ "prefresh" $
    prefresh pad (fromIntegral pminrow) (fromIntegral pmincol) (fromIntegral sminrow) (fromIntegral smincol) (fromIntegral smaxrow) (fromIntegral smaxcol)

delWin :: Window -> IO ()
delWin w = throwIfErr_ "delwin" $ delwin w
    
foreign import ccall unsafe prefresh :: Window -> CInt -> CInt -> CInt -> CInt -> CInt -> CInt -> IO CInt
foreign import ccall unsafe newpad :: CInt -> CInt -> IO Window
foreign import ccall unsafe delwin :: Window -> IO CInt


newWin :: Int -> Int -> Int -> Int -> IO Window
newWin nlines ncolumn begin_y begin_x = throwIfNull "newwin" $ newwin (fi nlines) (fi ncolumn) (fi begin_y) (fi begin_x)

foreign import ccall unsafe newwin :: CInt -> CInt -> CInt -> CInt -> IO Window


wClrToEol :: Window -> IO ()
wClrToEol w = throwIfErr_ "wclrtoeol" $ wclrtoeol w
foreign import ccall unsafe wclrtoeol :: Window -> IO CInt


getCh :: IO Key
getCh = threadWaitRead 0 >> (liftM decodeKey $ throwIfErr "getch" getch)

--getCh :: IO Key
--getCh = liftM decodeKey $ throwIfErr "getch" getch

-- getCh :: IO Key
-- getCh = do
--     nodelay stdScr 1
--     --halfdelay 1
--     v <- getch
--     case v of
-- 	(#const ERR) -> yield >> getCh 
-- 	x -> return $ decodeKey x

foreign import ccall unsafe halfdelay :: CInt -> IO CInt
foreign import ccall threadsafe getch :: IO CInt
foreign import ccall unsafe nodelay :: Window -> (Int8) -> IO CInt
{-# LINE 932 "Curses.hsc" #-}


foreign import ccall unsafe def_prog_mode :: IO CInt
foreign import ccall unsafe reset_prog_mode :: IO CInt

foreign import ccall "system" c_system :: Ptr CChar -> IO CInt

mySystem :: String -> IO Int 
mySystem s = fmap fromIntegral $ withCString s c_system
